/*
 * main.c
 *
 * Created: 21.5.2013 19:12:41
 *  Author: xkralj07
 */ 

#include <stdint.h>
#include <avr/io.h>					
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "lcd_driver.h"
#include "load.h"
#include "charger.h"
#include "btns_led.h"
#include "adc.h"
#include "math.h"

#define F_CPU				1000000									/* 8 MHz internal oscillator with 8 division */
#define TIMER1_PRESCALER	8
#define TIME_IN_T1_TICK(t)	((uint32_t)t*F_CPU/TIMER1_PRESCALER)	/* number of ticks of timer 1 corresponding to time t in seconds */
#define TIME_IN_T1_OVFS(t)	(TIME_IN_T1_TICK(t)/65536)				/* number of overflows of timer 1 corresponding to time t in seconds */

#define BTN_DEAD_PERIOD		TIME_IN_T1_TICK(0.004)					/* 4 ms in ticks of Timer 1 */

#define WAKING_TIMER1_OVF	0x01
#define WAKING_INT0			0x02
#define WAKING_INT1			0x04
#define WAKING_ADC			0x08
#define WAKING_COMPA		0x10
#define WAKING_COMPB		0x20

#define FCNS_ENA_LOAD		0x01
#define FCNS_ENA_CHARGER	0x02
#define FCNS_ENA_LCD		0x04
#define FCNS_ENA_OUT_LED	0x08
#define FCNS_CHAR_TIMER		0x10


#define INT_FLAG_LOAD_OFF	1
#define INT_FLAG_INT0_DEAD	2
#define INT_FLAG_INT1_DEAD	4

// constants correspond to 5V reference
#define MINIM_ACCU_VOLT		688						// 10.5 V
#define MINIM_CHAR_VOLT		539						// 15.0 V
#define MINIM_CURR_LIM_VOL	154						//  1.5 V

#define ACCU_V_FACT_POINT	22
//#define ACCU_V_FACT			uint16_t(0.015271775 << ACCU_V_FACT_POINT)
#define ACCU_V_FACT			64054

typedef enum
{
	ACCU_VOLTAGE,
	CHARGER_VOLTAGE,
	CHARGER_CURR_LIMITED,
} adc_state_t;


static volatile uint8_t waking_code = 0;				// binary coded flags that says what was the waking source of the processor
static volatile uint8_t fcns_ena = 0;					// binary coded flags of the enabled functions
static volatile uint8_t int_flags = 0;					// interrupt flags

// voltage measurement values
static uint16_t accu_volt = 0;
static uint8_t  char_volt = 0;


// Output button interrupt service routine
ISR(INT0_vect)
{
	waking_code |= WAKING_INT0;
	int_flags |= INT_FLAG_INT0_DEAD;

	OCR1A = TCNT1 + BTN_DEAD_PERIOD;					// set button dead interval
	TIFR1 |= _BV(OCIE1A);								// clear any pending interrupt for OCM
	TIMSK1 |= _BV(OCIE1A);								// enable interrupt for Timer 1 Output Compare Match
	BtnDisableInt0();									// disable button interrupt (it will be re-enabled after dead interval)

	// check interrupt sensitivity
	if (IsBtnSetInt0SensLowLevel())						// the button was pressed
	{
		// button pressed action is changing load state
		if (IsLoadOn())
		{
			// load is on, turn it off
			LoadOnDis();
			int_flags |= INT_FLAG_LOAD_OFF;
			fcns_ena &= ~FCNS_ENA_OUT_LED;
		}
		else if (fcns_ena & FCNS_ENA_LOAD)
		{
			// load is off and is enabled, turn it on
			LoadOnEna();
			fcns_ena |= FCNS_ENA_OUT_LED;
		}
	}
	else												// the button was released
	{
		
	}
	
	if (int_flags & INT_FLAG_LOAD_OFF)
	{
		int_flags &= ~INT_FLAG_LOAD_OFF;
		LoadOffEna();
	}
}

// LCD button interrupt service routine
ISR(INT1_vect)
{
	waking_code |= WAKING_INT1;
	int_flags |= INT_FLAG_INT1_DEAD;

	OCR1A = TCNT1 + BTN_DEAD_PERIOD;					// set button dead interval
	TIFR1 |= _BV(OCIE1A);								// clear any pending interrupt for OCM
	TIMSK1 |= _BV(OCIE1A);								// enable interrupt for Timer 1 Output Compare Match
	BtnDisableInt1();									// disable button interrupt (it will be re-enabled after dead interval)

	// check interrupt sensitivity
	if (IsBtnSetInt1SensLowLevel())						// the button was pressed
	{
		fcns_ena |= FCNS_ENA_LCD;						// set enable flag for LCD
	}
	else												// the button was released
	{
		
	}

}

// Timer 1 Output Compare Match A - for dead interval of buttons
ISR(TIMER1_COMPA_vect)
{
	waking_code |= WAKING_COMPA;
	TIMSK1 &= ~_BV(OCIE1A);								// disable interrupt for Timer 1 Output Compare Match

	if (int_flags & INT_FLAG_INT0_DEAD)					// if interrupt is end of the dead interval for INT 0
	{
		// change INT 0 sense
		if (IsBtnSetInt0SensLowLevel())
			BtnSetInt0SensRisingEdge();
		else
			BtnSetInt0SensLowLevel();
			
		BtnClearInt0();									// clear pending interrupt before re-enabling interrupt for button
	}
	
	if (int_flags & INT_FLAG_INT1_DEAD)					// if interrupt is end of the dead interval for INT 1
	{
		// change INT 1 sense
		if (IsBtnSetInt1SensLowLevel())
			BtnSetInt1SensRisingEdge();
		else
			BtnSetInt1SensLowLevel();
			
		BtnClearInt1();									// clear pending interrupt before re-enabling interrupt for button
	}
	
	int_flags &= ~(INT_FLAG_INT0_DEAD | INT_FLAG_INT1_DEAD);
	
	BtnEnableInt01();									// re-enable interrupts for both buttons
}

// Timer 1 Output Compare Match B - output LED turning off
ISR(TIMER1_COMPB_vect)
{
	waking_code |= WAKING_COMPB;
	
	TIMSK1 &= ~_BV(OCIE1B);								// disable interrupt for Timer 1 Output Compare Match
	LedOutSetOff();	
}

// Timer 1 overflow interrupt service routine with frequency = 2 Hz
ISR(TIMER1_OVF_vect)
{
	waking_code |= WAKING_TIMER1_OVF;
}

// ADC conversion completed service routine
ISR(ADC_vect)
{
	waking_code |= WAKING_ADC;
}

void LcdInfoShow(void)
{
	// constants in RAM, because actually there is more free space than in flash
	static char napeti_str [] = "Napeti ";
	static char aku_str [] = "aku=";
	uint32_t voltage;
	
	LcdTurnOn();
	LcdBacklightOn();
	
	LcdPrintString(napeti_str);
	LcdPrintString(aku_str);
	
	// print voltage of accumulator
	voltage = MultiplyUint(accu_volt, ACCU_V_FACT);
	LcdPrintFixedPoint((uint16_t)(voltage >> 16), ACCU_V_FACT_POINT - 16);
}

int main(void)
{
	uint8_t led_cnt = 0;
	uint8_t lcd_cnt = 0;
	uint16_t char_cnt = 0;
	
	adc_state_t adc_state = ACCU_VOLTAGE;
			
	BtnsLedInit();										// set input and output ports for buttons and output led, enable interrupts for falling edge of buttons inputs
	AdcInit();											// set prescaler and disable unused digital inputs
	LoadInit();											// set load driving pins as outputs
	LoadOffEna();										// switch transistor that turns off the load
	
	LcdInit();											// initialize LCD
	
	// Timer 1 settings
	TCCR1B =  _BV(CS11);								// set TIMER 1 - prescaler 1/8 (overflow frequency 1.91 Hz <=> period 524 ms)
	TIMSK1 |= _BV(TOIE1);								// enable timer 1 overflow interrupt
	
	PRR = _BV(PRTWI) | _BV(PRSPI) | _BV(PRTIM0);		// disable TWI, SPI and TIMER 0
	
	sleep_enable();										// sleep is enabled all the time
	
	sei();
	
	fcns_ena |= FCNS_ENA_LOAD;							// on startup load is enabled

	AdcDisable();										// disable ADC
	
    while(1)
    {
		set_sleep_mode(SLEEP_MODE_IDLE);				// set sleep mode idle
		sleep_bod_disable();							// disable brown-out detector during sleep
		sleep_cpu();
		
		LoadOffDis();									// maximum after 0.5 s disable load off signal
			
		// execute functions based on the waking code
		if (waking_code & WAKING_TIMER1_OVF)
		{
			waking_code &= ~WAKING_TIMER1_OVF;
			
			AdcEnable();									// enable adc
			AdcMuxSelAccuV();
			
			// if output LED is enabled, LED will be turned on for short time with period of blinking 2s
			if (fcns_ena & FCNS_ENA_OUT_LED)
			{
				led_cnt ++;
				if (led_cnt & 0x40)
				{
					LedOutSetOn();
					OCR1B = TCNT1 + 12500;					// short time of output LED shining is 100 ms
					TIFR1  |= _BV(OCIE1B);					// clear any pending interrupt
					TIMSK1 |= _BV(OCIE1B);					// enable interrupt of compare match B for LED turning off
				}
			}
			else
				led_cnt = 0;								// clear led counter to LED be start blinking when button is pressed
				
			// LCD is turned on only when it is enabled
			if (fcns_ena & FCNS_ENA_LCD)
			{
				if (lcd_cnt == 0)
					LcdInfoShow();
				
				if (lcd_cnt == 8)							// after 4s (approx.) LCD will be turned off
				{
					LcdPowerOff();							// power off LCD
					fcns_ena &= ~FCNS_ENA_LCD;				// disable LCD
					lcd_cnt = 0;
				}
				else
					lcd_cnt ++;
			}
			
			// charger timer
			if (fcns_ena & FCNS_CHAR_TIMER)
			{
				// count 1 hour to accumulator be charged with constant voltage
				
				if (char_cnt > TIME_IN_T1_OVFS(3600))
				{
					ChargerDisable();					// disable charger
				}
				else
				{
					char_cnt ++;
				}
			}
		}
		
		if (waking_code & WAKING_ADC)
		{
			waking_code &= ~WAKING_ADC;
			
			if (adc_state == ACCU_VOLTAGE)
			{
				accu_volt = AdcGetValue();
				
				if (accu_volt < MINIM_ACCU_VOLT)
				{
					fcns_ena &= ~(FCNS_ENA_LOAD | FCNS_ENA_OUT_LED);
					LoadOffEna();						// turn off load
					
					// TODO: the processor can sleep deeply now
				}
				
				adc_state = CHARGER_VOLTAGE;			// next state is charger voltage measurement
				AdcMuxSelCharV();
			}
			else if (adc_state == CHARGER_VOLTAGE)
			{
				char_volt = AdcGetValue() >> 2;			// to make sure the voltage is one byte length
				
				if (char_volt >= MINIM_CHAR_VOLT)
				{
					if (!(fcns_ena & FCNS_ENA_CHARGER))
					{
						if (char_cnt >= 4)				// valid power source is considered after 2s (approx.)
						{
							// just now charger was plug into power source
							// disable load and enable charger
							fcns_ena = (fcns_ena | FCNS_ENA_CHARGER) & ~(FCNS_ENA_LOAD | FCNS_ENA_OUT_LED);
							LoadOffEna();				// turn off load
							ChargerEnable();			// enable charger
							char_cnt = 0;				// zero charger counter
						}
						else
						{
							char_cnt ++;				// increment charger counter
						}
					}
				}
				else if (fcns_ena & FCNS_ENA_CHARGER)
				{
					// charger was unplug from power source just now
					// re-enable load and disable charger and charger timer
					fcns_ena = (fcns_ena | FCNS_ENA_LOAD) & ~(FCNS_ENA_CHARGER | FCNS_CHAR_TIMER);
					ChargerDisable();
					char_cnt = 0;
				}
				
				if (fcns_ena & FCNS_ENA_CHARGER)
				{
					// if charger is enabled, current limitation has to be determined
					adc_state = CHARGER_CURR_LIMITED;
					AdcMuxSelILimited();
				}
				else
				{
					// if charger is disabled, this is the end of ADC measurement, disable ADC and go to first state
					adc_state = ACCU_VOLTAGE;
					AdcMuxSelAccuV();
					AdcDisable();
				}
			}
			else if (adc_state == CHARGER_CURR_LIMITED)
			{
					if (AdcGetValue() < MINIM_CURR_LIM_VOL)
					{
						fcns_ena |= FCNS_CHAR_TIMER;
					}
					
					// this is the end of ADC measurement, disable ADC and go to first state
					adc_state = ACCU_VOLTAGE;
					AdcMuxSelAccuV();
					AdcDisable();				
			}
		}
	
		// unused waking codes
		/*
		if (waking_code & WAKING_INT0)
		{
			waking_code &= ~WAKING_INT0;
		}
				
		if (waking_code & WAKING_COMPA)
		{
			waking_code &= ~WAKING_COMPA;
		}
				
		if (waking_code & WAKING_COMPB)
		{
			waking_code &= ~WAKING_COMPB;
		}
		*/		
    }
}